﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace TIFile
{
    public class Program8XP
    {
        static List<byte> AllBytes;
        static StreamWriter writer;

        /// <summary>
        /// Builds a *.bin file into a *.8xp file
        /// </summary>
        /// <param name="bin">The *.bin file</param>
        /// <param name="ProgName">The name of the program on-calc</param>
        /// <param name="output">The output *.8xp file</param>
        public static void ConvertBinTo8XP(string bin, string ProgName, string output)
        {
            StreamReader reader = new StreamReader(bin);
            byte[] binBytes = new byte[reader.BaseStream.Length];
            reader.BaseStream.Read(binBytes, 0, (int)reader.BaseStream.Length);
            reader.Close();
            Program8XP.Save8XP(true, binBytes, ProgName, output);
        }

        /// <summary>
        /// Saves a *.8xp program
        /// </summary>
        /// <param name="Protected">Determines whether the program will be "locked" on-calc</param>
        /// <param name="Data">The program data, in bytes</param>
        /// <param name="ProgName">The string representation of the name</param>
        /// <param name="File">The file to save to</param>
        public static void Save8XP(bool Protected, byte[] Data, string ProgName, string File)
        {
            // Instatiate the byte list that will be used to create the checksum
            AllBytes = new List<byte>();
            int Length = Data.Length;
            // Fix the name of the program
            ProgName.ToUpper();
            if (ProgName.Length > 8)
                ProgName.Remove(8);
            Console.WriteLine("Creating program " + ProgName);
            // Create the stream
            writer = new StreamWriter(File);
            // The header (something to the effect of **TI83F*, with a bunch of zeroes)
            byte[] header = HexToBytes("2a2a54493833462a000000");
            // Manually fix some of the header bytes (I can't remember the hex equivalents of these and am too lazy to fix it)
            header[8] = 26;
            header[9] = 10;
            // Write our comment (I think it's "Created by Create8X" followed by zeroes, I calculated it yesterday)
            byte[] comment = HexToBytes("4372656174656420627920436f6e76657274385800000000000000000000000000000000000000000000");
            Console.WriteLine("Begin write...");
            // Write the header and comment
            Write(header, 0, header.Length);
            Write(comment, 0, comment.Length);
            // Theres a whole bunch of weird stuff about the length, don't ask me why
            Write(BitConverter.GetBytes((Int16)(Length + 19)), 0, 2);
            // Weird stuff
            Write(new byte[] { 13, 0 }, 0, 2);
            Write(BitConverter.GetBytes((Int16)(Length + 2)), 0, 2);
            // Write the protected value (6 = protected)
            WriteByte((byte)(Protected ? 6 : 5));
            // The name
            Write(StringToBytes(ProgName), 0, ProgName.Length);
            // The name must take up 8 characters.  If it doesn't, we have to add zeroes
            if (ProgName.Length < 8)
                for (int i = 0; i < 8 - ProgName.Length; i++)
                    WriteByte(0);
            // More weird stuff
            Write(new byte[] { 0, 0 }, 0, 2);
            // More length stuff
            Write(BitConverter.GetBytes((Int16)(Length + 2)), 0, 2);
            Write(BitConverter.GetBytes((Int16)(Length)), 0, 2);
            // The actual length, finally
            Write(Data, 0, Length);
            // And the checksum (all of the bytes from 36 forward, added together and truncated for overflow)
            writer.BaseStream.Write(GetChecksum(), 0, GetChecksum().Length);
            // And save all that hard work.
            writer.Close();
            Console.WriteLine("Write successful.");
        }

        #region Helpers
        private static void Write(byte[] bytes, int startIndex, int Length)
        {
            AllBytes.AddRange(bytes);
            writer.BaseStream.Write(bytes, startIndex, Length);
        }

        private static void WriteByte(byte b)
        {
            AllBytes.Add(b);
            writer.BaseStream.WriteByte(b);
        }

        private static byte[] GetChecksum()
        {
            short chk = 0;
            unchecked
            {
                for (int i = 36; i < AllBytes.Count; i++)
                    chk += (short)AllBytes[i];
            }
            return BitConverter.GetBytes(chk);
        }

        private static byte HexToByte(string HexValue)
        {
            int num = 0;
            int.TryParse(HexValue, System.Globalization.NumberStyles.HexNumber, null, out num);
            return (byte)num;
        }

        private static byte[] StringToBytes(string Str)
        {
            byte[] result = new byte[Str.Length];
            int i = 0;
            string start = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
            string convert = "4142434445464748494a4b4c4d4e4f505152535455565758595a";
            foreach (char c in Str)
            {
                result[i] = HexToByte(convert.Substring(start.IndexOf(c.ToString()) * 2, 2));
                i++;
            }
            return result;
        }

        public static byte[] HexToBytes(string Hex)
        {
            byte[] res = new byte[Hex.Length / 2];
            for (int i = 0; i < Hex.Length; i += 2)
            {
                res[i / 2] = HexToByte(Hex.Substring(i, 2));
            }
            return res;
        }
        #endregion
    }
}
